// ###############################################################################
// filename:		USB_MAIN.C
//
// Authors:		Reston Condit, Dan Butler
// Company:		Microchip Technology Inc
// Revision:	2.00 (Polling version)
// Date:		06 November 2002
//
// This file implements a basic interrupt service routine and shows how the
// USB interrupt should be serviced, how InitUSB should be called, and how PutUSB
// should be called.  It may be used as a reference, or as a starting point 
// from which to build an application.  
//
// ###############################################################################
//
// Stelladaptor: Atari 2600 Controller-to-USB Interface
// http://www.grandideastudio.com/portfolio/stelladaptor-2600/
//
// SW REV = 1.3, 10/19/15 (Implemented bug fix in USB_CH9.C, adjusted configuration bits, open source!)
// HW REV = A
//
//################################################################################

#include <pic.h>
#include <string.h>
#include "usb.h"
#include "usb_defs.h"

// Configuration Bits:	H4 Oscillator, WDT OFF, Power up timer ON, Code protection OFF
__CONFIG(WDTDIS & PWRTEN & H4 & UNPROTECT);

#define PORTBIT(adr, bit)   ((unsigned)(&adr)*8+(bit))

#define PACKET_SIZE     4

// joystick and driving controller
static bit UP       @ PORTBIT(PORTB, 4); // DB9 pin 1
static bit DOWN     @ PORTBIT(PORTB, 2); // DB9 pin 2
static bit LEFT     @ PORTBIT(PORTB, 1); // DB9 pin 3
static bit RIGHT    @ PORTBIT(PORTB, 0); // DB9 pin 4
static bit FIRE     @ PORTBIT(PORTB, 3); // DB9 pin 6

// paddle
static bit FIREB    @ PORTBIT(PORTB, 1); // DB9 pin 3
static bit FIREA    @ PORTBIT(PORTB, 0); // DB9 pin 4
static bit PADDLEA  @ PORTBIT(PORTA, 0); // DB9 pin 5
static bit PADDLEB  @ PORTBIT(PORTA, 1); // DB9 pin 9

static bit DEBUG    @ PORTBIT(PORTB, 7);

const char copyright[] = "Designed by Joe Grand, Grand Idea Studio [www.grandideastudio.com], CC BY 3.0 US 2015";


// Interrupt service routine.
interrupt ISR () 
{
    if (USBIF && USBIE) 
    {
		if (ACTIVITY)
			USBActivity ();
	
    	if (USB_RST && USB_RST_E) // USB reset must be serviced immediately
			USBReset ();
	
    	if (TOK_DONE && TOK_DONE_E) 
        {
			if (USB_dev_req == SET_ADDRESS)	// Finish Set Address
			{   				 
				USB_dev_req = NULL;
				USB_Curr_Config = 0;
				UADDR = USB_address_pending;
				UIE = 0x01;		// enable just the reset interrupt

 				if (USB_address_pending > 0)
					USWSTAT = ADDRESS_STATE;
				else
					USWSTAT = DEFAULT_STATE;
			}

			TOK_DONE = 0; // clear Token Done flag
		}

		USBIF = 0; // Clear USB interrupt flag
	}
}


void delay_us (int us) // 24MHz clock
{
    #asm
    delay1  nop 
            nop
            nop
            decfsz _delay_us$us
            goto delay1
    #endasm
}


unsigned char read_adc (unsigned char adc_channel)
{   
    CHS2 = (adc_channel & 0x04) >> 2;    // channel select
    CHS1 = (adc_channel & 0x02) >> 1;
    CHS0 = (adc_channel & 0x01);
    delay_us(20);   // wait the required acquisition time (16uS worst case)
    
    ADGO = 1;       // start conversion
    while(ADGO){};  // wait for a/d conversion to complete

    return (ADRESH);
}


void main() 
{
	unsigned char x, y;
	unsigned char buffer [PACKET_SIZE], previous[PACKET_SIZE];

    /*  report 
        byte 0: <reserved> data
        byte 1: x data (joystick left/right or paddle a)
        byte 2: y data (joystick up/down or paddle b)
        byte 3: buttons (padding[6] : fireb : fire/firea)
    */

    // Configure ports
	PORTB = 0;
	TRISB = 0x7F;   // set port b 0-6 as digital inputs, port b 7 as output (for debug)
    RBPU = 0;       // enable internal pull-ups

    PORTA = 0;
	TRISA = 0xFF;   // set port a as inputs
    ADCON1 = 0x04;  // AN0, AN1, AN3 analog inputs, all others digital, Vref = VDD;
    ADCON0 = 0xC1;  // Internal RC conversion clock, enable A/D converter module

    // USB initialization
	delay_us(100);	// Small delay (greater than 16us) to allow SIE to come online
	InitUSB ();

    // Timer initialization
	OPTION = 0x07;	// TMR0 prescaler 1:256

    // report initialization
    buffer[0] = 0; // reserved byte    

	while (1)
	{
	    if (T0IF) // Poll all functions every 10.6ms
        {				
			T0IF = 0;
			ServiceUSB();   // Service USB functions

            // Auto-detect for 2600 controller type
            x = read_adc(0);  // paddle a
            y = read_adc(1);  // paddle b

            // process joystick, driving controller, and other non-analog
            if ((x < 0x0F) && (y < 0x0F)) // set voltage threshold to 0.3V to handle floating voltages when no paddle connected
            {
                // X vector (0x00 to 0xFF, 0x7F = center)
                if (!LEFT && !RIGHT)
                    buffer[1] = 0xC0;   // must be a footpedal, track & field, or other controller, as no regular josystick can do this
                else if (!RIGHT)
                    buffer[1] = 0xFF;
                else if (!LEFT)
                    buffer[1] = 0x00;
                else 
                    buffer[1] = 0x7F;
                
                // Y vector (0x00 to 0xFF, 0x7F = center)
                if (!UP && !DOWN)
                    buffer[2] = 0xC0;   // must be a driving controller, as no regular joystick can do this
                else if (!DOWN)
                    buffer[2] = 0xFF;
                else if (!UP)
                    buffer[2] = 0x00;
                else
                    buffer[2] = 0x7F; 

                // fire button
                if (!FIRE)
                	buffer[3] = 0x01;
                else
                    buffer[3] = 0x00;
            }
            else    // process paddle            
            {
                // Normalize value for full range
                if (x < 0x80) x = 0x80; // Set minimum (floor) value
                if (y < 0x80) y = 0x80;
                buffer[1] = (x - 0x80) << 1;  // X vector (0x00 to 0xFE, 0x7E = center)
                buffer[2] = (y - 0x80) << 1;  // Y vector (0x00 to 0xFE, 0x7E = center)

                buffer[3] = 0;

                // fire buttons
                if (!FIREA)
                {
                	buffer[3] |= 0x01; // set button bit
                }
                else
                {
                    buffer[3] &= ~0x01; // clear button bit
                }
                if (!FIREB)
                {
                	buffer[3] |= 0x02;
                }
                else
                {
                    buffer[3] &= ~0x02;
                }
            }
           
            // Only send data if has changed
            if (memcmp(previous, buffer, PACKET_SIZE) != 0)
            {
                memcpy(previous, buffer, PACKET_SIZE);

                // Wait until device is configured before using
        		// EP1. If Endpoints 1 or 2 are used before
    			// the device is configured, errors will occur.
    			if (ConfiguredUSB()) 
                {		
    			   PutEP1(PACKET_SIZE, buffer);
    			}
            }
		}
	}
}
